@@ -9,10 +9,10 @@
 #
 import re
 import math
-from typing import Optional, SupportsFloat, SupportsInt, Union
+from typing import Any, Optional, SupportsFloat, SupportsInt, Union, Type
 
-from ..helpers import collapse_white_spaces
-from .atomic_types import AtomicTypeMeta, AnyAtomicType
+from elementpath.helpers import NUMERIC_INF_OR_NAN, INVALID_NUMERIC, collapse_white_spaces
+from .atomic_types import AnyAtomicType
 
 
 class Float10(float, AnyAtomicType):
@@ -25,89 +25,120 @@ class Float10(float, AnyAtomicType):
     def __new__(cls, value: Union[str, SupportsFloat]) -> 'Float10':
         if isinstance(value, str):
             value = collapse_white_spaces(value)
-            if value in {'INF', '-INF', 'NaN'} or cls.xsd_version != '1.0' and value == '+INF':
-                pass
-            elif value.lower() in {'inf', '+inf', '-inf', 'nan',
-                                   'infinity', '+infinity', '-infinity'}:
+            if value in NUMERIC_INF_OR_NAN or cls.xsd_version != '1.0' and value == '+INF':
+                if value == 'NaN':
+                    try:
+                        return float_nan
+                    except NameError:
+                        pass
+            elif value.lower() in INVALID_NUMERIC:
                 raise ValueError('invalid value {!r} for xs:{}'.format(value, cls.name))
+        elif math.isnan(value):
+            try:
+                return float_nan
+            except NameError:  # pragma: no cover
+                pass
 
-        value = super().__new__(cls, value)
-        if value > 3.4028235E38:
+        _value = super().__new__(cls, value)
+        if _value > 3.4028235E38:
             return super().__new__(cls, 'INF')
-        elif value < -3.4028235E38:
+        elif _value < -3.4028235E38:
             return super().__new__(cls, '-INF')
-        elif -1e-37 < value < 1e-37:
-            return super().__new__(cls, -0.0 if str(value).startswith('-') else 0.0)
-        return value
+        elif -1e-37 < _value < 1e-37:
+            return super().__new__(cls, -0.0 if str(_value).startswith('-') else 0.0)
+        return _value
 
-    def __hash__(self):
+    def __init__(self, value: Union[str, SupportsFloat]) -> None:
+        float.__init__(self)
+
+    def __hash__(self) -> int:
         return super(Float10, self).__hash__()
 
-    def __eq__(self, other):
+    def __eq__(self, other: object) -> bool:
         if isinstance(other, self.__class__):
             if super(Float10, self).__eq__(other):
                 return True
             return math.isclose(self, other, rel_tol=1e-7, abs_tol=0.0)
         return super(Float10, self).__eq__(other)
 
-    def __ne__(self, other):
+    def __ne__(self, other: object) -> bool:
         if isinstance(other, self.__class__):
             if super(Float10, self).__eq__(other):
                 return False
             return not math.isclose(self, other, rel_tol=1e-7, abs_tol=0.0)
         return super(Float10, self).__ne__(other)
 
-    def __add__(self, other):
-        if isinstance(other, (self.__class__, int)):
+    def __add__(self, other: object) -> Union[float, 'Float10', 'Float']:
+        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
             return self.__class__(super(Float10, self).__add__(other))
-        return super(Float10, self).__add__(other)
+        elif isinstance(other, float):
+            return super(Float10, self).__add__(other)
+        return NotImplemented
 
-    def __radd__(self, other):
-        if isinstance(other, (self.__class__, int)):
+    def __radd__(self, other: object) -> Union[float, 'Float10', 'Float']:
+        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
             return self.__class__(super(Float10, self).__radd__(other))
-        return super(Float10, self).__radd__(other)
+        elif isinstance(other, float):
+            return super(Float10, self).__radd__(other)
+        return NotImplemented
 
-    def __sub__(self, other):
-        if isinstance(other, (self.__class__, int)):
+    def __sub__(self, other: object) -> Union[float, 'Float10', 'Float']:
+        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
             return self.__class__(super(Float10, self).__sub__(other))
-        return super(Float10, self).__sub__(other)
+        elif isinstance(other, float):
+            return super(Float10, self).__sub__(other)
+        return NotImplemented
 
-    def __rsub__(self, other):
-        if isinstance(other, (self.__class__, int)):
+    def __rsub__(self, other: object) -> Union[float, 'Float10', 'Float']:
+        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
             return self.__class__(super(Float10, self).__rsub__(other))
-        return super(Float10, self).__rsub__(other)
+        elif isinstance(other, float):
+            return super(Float10, self).__rsub__(other)
+        return NotImplemented
 
-    def __mul__(self, other):
-        if isinstance(other, (self.__class__, int)):
+    def __mul__(self, other: object) -> Union[float, 'Float10', 'Float']:
+        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
             return self.__class__(super(Float10, self).__mul__(other))
-        return super(Float10, self).__mul__(other)
+        elif isinstance(other, float):
+            return super(Float10, self).__mul__(other)
+        return NotImplemented
 
-    def __rmul__(self, other):
-        if isinstance(other, (self.__class__, int)):
+    def __rmul__(self, other: object) -> Union[float, 'Float10', 'Float']:
+        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
             return self.__class__(super(Float10, self).__rmul__(other))
-        return super(Float10, self).__rmul__(other)
+        elif isinstance(other, float):
+            return super(Float10, self).__rmul__(other)
+        return NotImplemented
 
-    def __truediv__(self, other):
-        if isinstance(other, (self.__class__, int)):
+    def __truediv__(self, other: object) -> Union[float, 'Float10', 'Float']:
+        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
             return self.__class__(super(Float10, self).__truediv__(other))
-        return super(Float10, self).__truediv__(other)
+        elif isinstance(other, float):
+            return super(Float10, self).__truediv__(other)
+        return NotImplemented
 
-    def __rtruediv__(self, other):
-        if isinstance(other, (self.__class__, int)):
+    def __rtruediv__(self, other: object) -> Union[float, 'Float10', 'Float']:
+        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
             return self.__class__(super(Float10, self).__rtruediv__(other))
-        return super(Float10, self).__rtruediv__(other)
+        elif isinstance(other, float):
+            return super(Float10, self).__rtruediv__(other)
+        return NotImplemented
 
-    def __mod__(self, other):
-        if isinstance(other, (self.__class__, int)):
+    def __mod__(self, other: object) -> Union[float, 'Float10', 'Float']:
+        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
             return self.__class__(super(Float10, self).__mod__(other))
-        return super(Float10, self).__mod__(other)
+        elif isinstance(other, float):
+            return super(Float10, self).__mod__(other)
+        return NotImplemented
 
-    def __rmod__(self, other):
-        if isinstance(other, (self.__class__, int)):
+    def __rmod__(self, other: object) -> Union[float, 'Float10', 'Float']:
+        if isinstance(other, (self.__class__, int)) and not isinstance(other, bool):
             return self.__class__(super(Float10, self).__rmod__(other))
-        return super(Float10, self).__rmod__(other)
+        elif isinstance(other, float):
+            return super(Float10, self).__rmod__(other)
+        return NotImplemented
 
-    def __abs__(self):
+    def __abs__(self) -> Union['Float10', 'Float']:
         return self.__class__(super(Float10, self).__abs__())
 
 
@@ -116,7 +147,11 @@ class Float(Float10):
     xsd_version = '1.1'
 
 
-class Integer(int, metaclass=AtomicTypeMeta):
+# The instance used for xs:float NaN values in order to keep identity
+float_nan = Float10('NaN')
+
+
+class Integer(int, AnyAtomicType):
     """A wrapper for emulating xs:integer and limited integer types."""
     name = 'integer'
     pattern = re.compile(r'^[\-+]?[0-9]+$')
@@ -128,22 +163,23 @@ class Integer(int, metaclass=AtomicTypeMeta):
             raise ValueError("value {} is too low for {!r}".format(value, self.__class__))
         elif self.higher_bound is not None and self >= self.higher_bound:
             raise ValueError("value {} is too high for {!r}".format(value, self.__class__))
-        super(Integer, self).__init__()
+        int.__init__(self)
 
     @classmethod
-    def __subclasshook__(cls, subclass):
+    def __subclasshook__(cls, subclass: Type[Any]) -> bool:
         if cls is Integer:
             return issubclass(subclass, int) and not issubclass(subclass, bool)
         return NotImplemented
 
     @classmethod
-    def validate(cls, value):
+    def validate(cls, value: object) -> None:
         if isinstance(value, cls):
             return
-        elif not isinstance(value, str):
+        elif isinstance(value, str):
+            if cls.pattern.match(value) is None:
+                raise cls.invalid_value(value)
+        else:
             raise cls.invalid_type(value)
-        elif cls.pattern.match(value) is None:
-            raise cls.invalid_value(value)
 
 
 class NonPositiveInteger(Integer):
