@@ -8,8 +8,14 @@
 # @author Davide Brunato <brunato@sissa.it>
 #
 import locale
-from typing import Optional
-from .tdop import Token
+from typing import TYPE_CHECKING, Any, Optional, Union
+
+if TYPE_CHECKING:
+    from elementpath.tdop import Token
+
+from elementpath.aliases import AnyNsmapType
+from elementpath.namespaces import XQT_ERRORS_NAMESPACE
+from elementpath import datatypes
 
 
 class ElementPathError(Exception):
@@ -22,7 +28,7 @@ class ElementPathError(Exception):
     """
     def __init__(self, message: str,
                  code: Optional[str] = None,
-                 token: Optional[Token] = None) -> None:
+                 token: Optional['Token[Any]'] = None) -> None:
         super(ElementPathError, self).__init__(message)
         self.message = message
         self.code = code
@@ -83,7 +89,7 @@ class ElementPathLocaleError(ElementPathError, locale.Error):
 
 
 XPATH_ERROR_CODES = {
-    # XPath 2.0 parser error (https://www.w3.org/TR/xpath20/#id-errors)
+    # XPath 2.0 parser errors (https://www.w3.org/TR/xpath20/#id-errors)
     'XPST0001': (ElementPathValueError, 'Parser not bound to a schema'),
     'XPST0003': (ElementPathSyntaxError, 'Invalid XPath expression'),
     'XPDY0002': (MissingContextError, 'Dynamic context required for evaluate'),
@@ -102,7 +108,7 @@ XPATH_ERROR_CODES = {
                  'Target type cannot be xs:NOTATION or xs:anyAtomicType'),
     'XPST0081': (ElementPathNameError, 'Unknown namespace'),
 
-    # XPath data types and function errors
+    # Data types and functions errors
     'FOER0000': (ElementPathError, 'Unidentified error'),
     'FOAR0001': (ElementPathZeroDivisionError, 'Division by zero'),
     'FOAR0002': (ElementPathOverflowError, 'Numeric operation overflow/underflow'),
@@ -115,7 +121,7 @@ XPATH_ERROR_CODES = {
     'FOCH0001': (ElementPathValueError, 'Code point not valid'),
     'FOCH0002': (ElementPathLocaleError, 'Unsupported collation'),
     'FOCH0003': (ElementPathValueError, 'Unsupported normalization form'),
-    'FOCH0004': (ElementPathValueError, 'Collation does not support collation units'),
+    'FOCH0004': (ElementPathLocaleError, 'Collation does not support collation units'),
     'FODC0001': (ElementPathValueError, 'No context document'),
     'FODC0002': (ElementPathValueError, 'Error retrieving resource'),
     'FODC0003': (ElementPathValueError, 'Function stability not defined'),
@@ -145,7 +151,11 @@ XPATH_ERROR_CODES = {
     'FORX0004': (ElementPathValueError, 'Invalid replacement string'),
     'FOTY0012': (ElementPathValueError, 'Argument node does not have a typed value'),
 
-    # XPath 3.0 errors
+    # XPath 3.0+ errors
+    'XQST0039': (ElementPathTypeError, 'Duplicate parameter name in inline function expression'),
+    'XQST0046': (ElementPathTypeError, 'The namespace part of the EQName is not a valid URI'),
+    'XQST0052': (ElementPathNameError, 'The name of an in-scope simple schema type required'),
+    'XQST0070': (ElementPathNameError, 'Illegal use of a predefined namespace'),
     'FOTY0013': (ElementPathTypeError, 'The argument to fn:data() contains a function item'),
     'FOTY0014': (ElementPathTypeError, 'The argument to fn:string() is a function item'),
     'FOTY0015': (ElementPathTypeError,
@@ -164,54 +174,132 @@ XPATH_ERROR_CODES = {
     'FOFD1340': (ElementPathValueError, 'Invalid date/time formatting parameters'),
     'FOFD1350': (ElementPathValueError, 'Invalid date/time formatting component'),
 
+    'XPTY0117': (ElementPathTypeError,
+                 'Item type is xs:untypedAtomic and the expected type is namespace-sensitive'),
+    'XPDY0130': (ElementPathValueError,
+                 'An implementation-defined limit has been exceeded'),
+    'XPST0133': (ElementPathValueError,
+                 'The namespace URI for EQName is http://www.w3.org/2000/xmlns/'),
+
     # XSLT and XQuery Serialization errors
     # (the complete list: https://www.w3.org/TR/xslt-xquery-serialization/#id-errors)
     'SENR0001': (ElementPathTypeError, 'item is an attribute node or a namespace node'),
     'SEPM0016': (ElementPathValueError, 'parameter value is invalid for the defined domain'),
     'SEPM0017': (ElementPathValueError, 'error during extraction of serialization parameters'),
-    'SEPM0018': (ElementPathValueError, 'use-character-maps serialization parameter in '
-                                        'a sequence of length greater than one'),
+    'SEPM0018': (ElementPathTypeError, 'use-character-maps serialization parameter in '
+                                       'a sequence of length greater than one'),
     'SEPM0019': (ElementPathValueError, 'same serialization parameter appears more than once'),
+    'SERE0020': (ElementPathTypeError, 'a numeric value being serialized using the JSON output '
+                                       'method cannot be represented in the JSON grammar'),
+    'SERE0021': (ElementPathTypeError, 'a sequence being serialized using the JSON output '
+                                       'method includes items for which no rules are provided '
+                                       'in the appropriate section of the serialization rules'),
+    'SERE0022': (ElementPathValueError, 'a map being serialized using the JSON output method '
+                                        'has two keys with the same string value'),
+    'SERE0023': (ElementPathTypeError, 'a sequence being serialized using the JSON output '
+                                       'method is of length greater than one'),
+
+    # XPath 3.1+ errors
+    'FOJS0001': (ElementPathSyntaxError, 'JSON syntax error'),
+    'FOJS0003': (ElementPathValueError, 'JSON duplicate keys'),
+    'FOJS0004': (ElementPathRuntimeError, 'JSON: not schema-aware'),
+    'FOJS0005': (ElementPathValueError, 'Invalid options'),
+    'FOJS0006': (ElementPathValueError, 'Invalid XML representation of JSON'),
+    'FOJS0007': (ElementPathValueError, 'Bad JSON escape sequence'),
+    'FOAY0001': (ElementPathValueError, 'Array index out of bounds'),
+    'FOAY0002': (ElementPathValueError, 'Negative array length'),
+
+    'FOQM0001': (ElementPathValueError, 'Module URI is a zero-length string'),
+    'FOQM0002': (ElementPathRuntimeError, 'Module URI not found'),
+    'FOQM0003': (ElementPathRuntimeError, 'Static error in dynamically-loaded XQuery module'),
+    'FOQM0005': (ElementPathValueError, 'Parameter for dynamically-loaded '
+                                        'XQuery module has incorrect type'),
+    'FOQM0006': (ElementPathRuntimeError, 'No suitable XQuery processor available'),
+
+    'FOXT0001': (ElementPathRuntimeError, 'No suitable XSLT processor available'),
+    'FOXT0002': (ElementPathValueError, 'Invalid parameters to XSLT transformation'),
+    'FOXT0003': (ElementPathRuntimeError, 'XSLT transformation failed'),
+    'FOXT0004': (ElementPathRuntimeError, 'XSLT transformation has been disabled'),
+    'FOXT0006': (ElementPathValueError, 'XSLT output contains non-accepted characters'),
+
+    'FOAP0001': (ElementPathTypeError, 'Wrong number of arguments'),
+    'FORG0010': (ElementPathValueError, 'Invalid date/time'),
+    'XQDY0137': (ElementPathValueError, 'No two keys in a map may have the same key value'),
 }
 
 
-def xpath_error(code: str, message: Optional[str] = None,
-                token: Optional[Token] = None, prefix: str = 'err') -> ElementPathError:
+def xpath_error(code: Union[str, 'datatypes.QName'],
+                message_or_error:  Union[None, str, Exception] = None,
+                token: Optional['Token[Any]'] = None,
+                namespaces: AnyNsmapType = None) -> ElementPathError:
     """
     Returns an XPath error instance related with a code. An XPath/XQuery/XSLT error code
     (ref: http://www.w3.org/2005/xqt-errors) is an alphanumeric token starting with four
     uppercase letters and ending with four digits.
 
     :param code: the error code.
-    :param message: an optional custom additional message.
+    :param message_or_error: an optional custom message or related exception.
     :param token: an optional token instance.
-    :param prefix: the namespace prefix to apply to the error code, defaults to 'err'.
+    :param namespaces: an optional namespace mapping for finding the prefix \
+    related with the namespace 'http://www.w3.org/2005/xqt-errors'.
+    For default the prefix 'err' is used.
     """
-    if code.startswith('{'):
-        try:
-            namespace, code = code[1:].split('}')
-        except ValueError:
-            message = '{!r} is not an xs:QName'.format(code)
-            raise ElementPathValueError(message, 'err:XPTY0004', token)
+    if isinstance(code, datatypes.QName):
+        namespace = code.uri
+        if namespace:
+            pcode, code = code.qname, code.local_name
         else:
-            if namespace != 'http://www.w3.org/2005/xqt-errors':
-                message = 'invalid namespace {!r}'.format(namespace)
-                raise ElementPathValueError(message, 'err:XPTY0004', token)
-            pcode = '%s:%s' % (prefix, code) if prefix else code
-    elif ':' not in code:
-        pcode = '%s:%s' % (prefix, code) if prefix else code
-    elif not prefix or not code.startswith(prefix + ':'):
-        message = '%r is not an XPath error code' % code
-        raise ElementPathValueError(message, 'err:XPTY0004', token)
+            pcode, code = code.braced_uri_name, code.local_name
     else:
-        pcode = code
-        code = code[len(prefix) + 1:]
+        namespace = XQT_ERRORS_NAMESPACE
+        prefix: Optional[str]
+        if not namespaces or namespaces.get('err') == XQT_ERRORS_NAMESPACE:
+            prefix = 'err'
+        else:
+            for prefix, uri in namespaces.items():
+                if uri == XQT_ERRORS_NAMESPACE:
+                    break
+            else:
+                prefix = 'err'
+
+        if code.startswith('{'):
+            try:
+                namespace, code = code[1:].split('}')
+            except ValueError:
+                message = '{!r} is not an xs:QName'.format(code)
+                raise ElementPathValueError(message, 'err:XPTY0004', token)
+            else:
+                pcode = f'{prefix}:{code}' if prefix else code
+
+        elif ':' not in code:
+            pcode = f'{prefix}:{code}' if prefix else code
+        elif code.startswith(f'{prefix}:') and code.count(':') == 1:
+            pcode, code = code, code.split(':')[1]
+        else:
+            message = '%r is not an XPath error code' % code
+            raise ElementPathValueError(message, 'err:XPTY0004', token)
+
+        if namespace != XQT_ERRORS_NAMESPACE:
+            message = 'invalid namespace {!r}'.format(namespace)
+            raise ElementPathValueError(message, 'err:XPTY0004', token)
 
     try:
         error_class, default_message = XPATH_ERROR_CODES[code]
     except KeyError:
-        raise ElementPathValueError(
-            message or 'unknown XPath error code %r' % code, 'err:XPTY0004', token
-        )
+        if namespace == XQT_ERRORS_NAMESPACE:
+            message = f'unknown XPath error code {code}'
+            raise ElementPathValueError(message, 'err:XPTY0004', token) from None
+        else:
+            error_class = ElementPathError
+            default_message = 'custom XPath error'
+
+    if message_or_error is None:
+        message = default_message
+    elif isinstance(message_or_error, str):
+        message = message_or_error
+    elif isinstance(message_or_error, ElementPathError):
+        message = message_or_error.message
     else:
-        return error_class(message or default_message, pcode, token)
+        message = str(message_or_error)
+
+    return error_class(message, pcode, token)
